
import json
import logging
from dataclasses import dataclass
from datetime import datetime

            
import aiohttp
import validators
from discord import Embed, Colour
from bs4 import BeautifulSoup as bs4
from feedparser import FeedParserDict, parse
from markdownify import markdownify
from sqlalchemy import select, insert, delete, and_
from sqlalchemy.exc import NoResultFound
from textwrap import shorten

from errors import IllegalFeed
from db import DatabaseManager, RssSourceModel, FeedChannelModel
from utils import get_rss_data, get_unparsed_feed

log = logging.getLogger(__name__)
dumps = lambda _dict: json.dumps(_dict, indent=8)


@dataclass
class Article:
    """Represents a news article, or entry from an RSS feed."""

    title: str | None
    description: str | None
    url: str | None
    published: datetime | None
    author: str | None
    source: object

    @classmethod
    def from_entry(cls, source, entry:FeedParserDict):
        """Create an Article from an RSS feed entry.

        Parameters
        ----------
        entry : FeedParserDict
            An entry pulled from a complete FeedParserDict object.

        Returns
        -------
        Article
            The Article created from the feed entry.
        """

        log.debug("Creating Article from entry: %s", dumps(entry))

        published_parsed = entry.get("published_parsed")
        published = datetime(*entry.published_parsed[0:-2]) if published_parsed else None

        return cls(
            title=entry.get("title"),
            description=entry.get("description"),
            url=entry.get("link"),
            published=published,
            author = entry.get("author"),
            source=source
        )

    async def get_thumbnail_url(self, session: aiohttp.ClientSession) -> str | None:
        """Returns the thumbnail URL for an article.

        Parameters
        ----------
        session : aiohttp.ClientSession
            A client session used to get the thumbnail.

        Returns
        -------
        str or None
            The thumbnail URL, or None if not found.
        """

        log.debug("Fetching thumbnail for article: %s", self)

        try:
            async with session.get(self.url, timeout=15) as response:
                html = await response.text()
        except aiohttp.InvalidURL as error:
            log.error("invalid thumbnail url: %s", error)
            return None

        soup = bs4(html, "html.parser")
        image_element = soup.select_one("meta[property='og:image']")
        if not image_element:
            return None

        image_content = image_element.get("content")
        return image_content if validators.url(image_content) else None

    async def to_embed(self, session: aiohttp.ClientSession) -> Embed:
        """Creates and returns a Discord Embed object from the article.

        Parameters
        ----------
        session : aiohttp.ClientSession
            A client session used to get additional article data.

        Returns
        -------
        Embed
            A Discord Embed object representing the article.
        """

        log.debug(f"Creating embed from article: {self}")

        # Replace HTML with Markdown, and shorten text.
        title = shorten(markdownify(self.title, strip=["img", "a"]), 256)
        desc = shorten(markdownify(self.description, strip=["img"]), 4096)
        author = shorten(self.source.name, 256)

        # validate urls
        embed_url = self.url if validators.url(self.url) else None
        author_url = self.source.url if validators.url(self.source.url) else None
        icon_url = self.source.icon_url if validators.url(self.source.icon_url) else None
        thumb_url = await self.get_thumbnail_url(session)  # validation done inside func

        embed = Embed(
            title=title,
            description=desc,
            timestamp=self.published,
            url=embed_url,
            colour=Colour.blue()
        )

        embed.set_thumbnail(url=icon_url)
        embed.set_image(url=thumb_url)
        embed.set_author(url=author_url, name=author)
        embed.set_footer(text=self.author)

        return embed


@dataclass
class Source:
    """Represents an RSS source."""
    
    name: str | None
    url: str | None
    icon_url: str | None
    feed: FeedParserDict

    @classmethod
    def from_parsed(cls, feed:FeedParserDict):
        """Returns a Source object from a parsed feed.

        Parameters
        ----------
        feed : FeedParserDict
            The feed used to create the Source.

        Returns
        -------
        Source
            The Source object
        """

        log.debug("Creating Source from feed: %s", dumps(feed))

        return cls(
            name=feed.get("channel", {}).get("title"),
            url=feed.get("channel", {}).get("link"),
            icon_url=feed.get("feed", {}).get("image", {}).get("href"),
            feed=feed
        )

    @classmethod
    async def from_url(cls, url: str):
        unparsed_content = await get_unparsed_feed(url)
        return cls.from_parsed(parse(unparsed_content))

    def get_latest_articles(self, max: int = 999) -> list[Article]:
        """Returns a list of Article objects.

        Parameters
        ----------
        max : int
            The maximum number of articles to return.

        Returns
        -------
        list of Article
            A list of Article objects.
        """

        log.debug("Fetching latest articles from %s, max=%s", self, max)

        return [
            Article.from_entry(self, entry)
            for i, entry in enumerate(self.feed.entries)
            if i < max
        ]


class Functions:

    def __init__(self, bot):
        self.bot = bot

    async def validate_feed(self, nickname: str, url: str) -> FeedParserDict:
        """Validates a feed based on the given nickname and url.

        Parameters
        ----------
        nickname : str
            Human readable nickname used to refer to the feed.
        url : str
            URL to fetch content from the feed.

        Returns
        -------
        FeedParserDict
            A Parsed Dictionary of the feed.

        Raises
        ------
        IllegalFeed
            If the feed is invalid.
        """

        # Ensure the URL is valid
        if not validators.url(url):
            raise IllegalFeed(f"The URL you have entered is malformed or invalid:\n`{url=}`")

        # Check the nickname is not a URL
        if validators.url(nickname):
            raise IllegalFeed(
                "It looks like the nickname you have entered is a URL.\n" \
                "For security reasons, this is not allowed.",
                nickname=nickname
            )

        feed_data, status_code = await get_rss_data(url)

        if status_code != 200:
            raise IllegalFeed(
                "The URL provided returned an invalid status code:",
                url=url, status_code=status_code
            )

        # Check the contents is actually an RSS feed.
        feed = parse(feed_data)
        if not feed.version:
            raise IllegalFeed(
                "The provided URL does not seem to be a valid RSS feed.",
                url=url
            )

        return feed

    async def create_new_feed(self, nickname: str, url: str, guild_id: int) -> Source:
        """Create a new Feed, and return it as a Source object.

        Parameters
        ----------
        nickname : str
            Human readable nickname used to refer to the feed.
        url : str
            URL to fetch content from the feed.
        guild_id : int
            Discord Server ID associated with the feed.

        Returns
        -------
        Source
            Dataclass containing attributes of the feed.
        """

        log.info("Creating new Feed: %s - %s", nickname, guild_id)

        parsed_feed = await self.validate_feed(nickname, url)

        async with DatabaseManager() as database:
            query = insert(RssSourceModel).values(
                discord_server_id=guild_id,
                rss_url=url,
                nick=nickname
            )
            await database.session.execute(query)

        log.info("Created Feed: %s - %s", nickname, guild_id)

        return Source.from_parsed(parsed_feed)

    async def delete_feed(self, url: str, guild_id: int) -> Source:
        """Delete an existing Feed, then return it as a Source object.

        Parameters
        ----------
        url : str
            URL of the feed, used in the whereclause.
        guild_id : int
            Discord Server ID of the feed, used in the whereclause.

        Returns
        -------
        Source
            Dataclass containing attributes of the feed.
        """

        log.info("Deleting Feed: %s - %s", url, guild_id)

        async with DatabaseManager() as database:
            whereclause = and_(
                RssSourceModel.discord_server_id == guild_id,
                RssSourceModel.rss_url == url
            )

            # Select the Feed entry, because an exception is raised if not found.
            select_query = select(RssSourceModel).filter(whereclause)
            select_result = await database.session.execute(select_query)
            select_result.scalars().one()

            delete_query = delete(RssSourceModel).filter(whereclause)
            await database.session.execute(delete_query)

        log.info("Deleted Feed: %s - %s", url, guild_id)

        return await Source.from_url(url)

    async def get_feeds(self, guild_id: int) -> list[tuple[str, str]]:
        """Returns a list of fetched Feed objects from the database.
        Note: a request will be made too all found Feed UR Ls.

        Parameters
        ----------
        guild_id : int
            The Discord Server ID, used to filter down the Feed query.

        Returns
        -------
        list[tuple[str, str]]
            List of Source objects, resulting from the query.

        Raises
        ------
        NoResultFound
            Raised if no results are found.
        """

        async with DatabaseManager() as database:
            whereclause = and_(RssSourceModel.discord_server_id == guild_id)
            query = select(RssSourceModel).where(whereclause)
            result = await database.session.execute(query)
            rss_sources = result.scalars().all()

            if not rss_sources:
                raise NoResultFound

            return [(feed.nick, feed.rss_url) for feed in rss_sources]

    async def assign_feed(
        self, url: str, channel_name: str, channel_id: int, guild_id: int
    ) -> tuple[int, Source]:
        """"""

        async with DatabaseManager() as database:
            select_query = select(RssSourceModel).where(and_(
                RssSourceModel.rss_url == url,
                RssSourceModel.discord_server_id == guild_id
            ))
            select_result = await database.session.execute(select_query)

            rss_source = select_result.scalars().one()

            insert_query = insert(FeedChannelModel).values(
                discord_server_id = guild_id,
                discord_channel_id = channel_id,
                rss_source_id=rss_source.id,
                search_name=f"{rss_source.nick} #{channel_name}"
            )

            insert_result = await database.session.execute(insert_query)
            return insert_result.inserted_primary_key.id, await Source.from_url(url)

    async def unassign_feed( self, assigned_feed_id: int, guild_id: int):
        """"""

        async with DatabaseManager() as database:
            query = delete(FeedChannelModel).where(and_(
                FeedChannelModel.id == assigned_feed_id,
                FeedChannelModel.discord_server_id == guild_id
            ))

            result = await database.session.execute(query)
            if not result.rowcount:
                raise NoResultFound
